/*
 * ggcov - A GTK frontend for exploring gcov coverage data
 * Copyright (c) 2001-2003 Greg Banks <gnb@alphalink.com.au>
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "cov.H"
#include "estring.H"
#include "filename.h"

CVSID("$Id: cov_arc.C,v 1.4 2003/06/06 15:21:30 gnb Exp $");

/*-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-*/

cov_arc_t::cov_arc_t(cov_block_t *from, cov_block_t *to)
{
    idx_ = from->out_arcs_.length();
    
    from_ = from;
    from_->out_arcs_.append(this);
    from_->out_ninvalid_++;

    to_ = to;
    to_->in_arcs_.append(this);
    to_->in_ninvalid_++;
}

cov_arc_t::~cov_arc_t()
{
    to_->in_arcs_.remove(this);
    from_->out_arcs_.remove(this);
    /* TODO: ninvalid counts?!?! */
}

/*-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-*/
/*
 * Return TRUE iff the arc represents a call to a different function
 * rather than a branch within the same function.
 *
 * We used to be able to detect function calls with the fake_ flag
 * from the .bbg file, but that flag went away in braindead versions
 * of gcc 2.96 so we have to fall back to a more subtle test.  The
 * test relies on the last block in a function being a fake block
 * whose purpose is to be the target of all call arcs in the function.
 * This works with most of the data generated by gcc 2.96, but sometimes
 * it screws that up too.  Note that the commandline "gcov" utility
 * that comes with gcc 2.96 cannot get any call stats at all.
 */

gboolean 
cov_arc_t::is_call() const
{
#ifdef HAVE_BBG_FAKE_FLAG
    return (fake_);
#else
    return (to_->bindex() == to_->function()->num_blocks()-1);
#endif
}

/*-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-*/

void
cov_arc_t::set_count(count_t count)
{
    assert(!count_valid_);
    count_valid_ = TRUE;
    count_ = count;
    from_->out_ninvalid_--;
    to_->in_ninvalid_--;
}

count_t
cov_arc_t::total(const list_t<cov_arc_t> &list)
{
    count_t total = 0;
    list_iterator_t<cov_arc_t> iter;
    
    for (iter = list.first() ; iter != (cov_arc_t *)0 ; ++iter)
    {
	/* Some of the counts will be invalid, but they are zero,
	   so adding it in also doesn't hurt.  */
	total += (*iter)->count_;
    }
    return total;
}

#ifdef HAVE_BBG_FAKE_FLAG
unsigned int
cov_arc_t::nfake(const list_t<cov_arc_t> &list)
{
    unsigned int nfake = 0;
    list_iterator_t<cov_arc_t> iter;

    for (iter = list.first() ; iter != (cov_arc_t *)0 ; ++iter)
    {
	if ((*iter)->fake_)
	    nfake++;
    }
    return nfake;
}	    
#endif

unsigned int
cov_arc_t::ncalls(const list_t<cov_arc_t> &list)
{
    unsigned int n = 0;
    list_iterator_t<cov_arc_t> iter;

    for (iter = list.first() ; iter != (cov_arc_t *)0 ; ++iter)
    {
	if ((*iter)->is_call())
	    n++;
    }
    return n;
}	    

cov_arc_t *
cov_arc_t::find_invalid(const list_t<cov_arc_t> &list)
{
    list_iterator_t<cov_arc_t> iter;

    for (iter = list.first() ; iter != (cov_arc_t *)0 ; ++iter)
    {
    	cov_arc_t *a = *iter;
	
	if (!a->count_valid_)
	    return a;
    }
    return 0;
}

/*-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-*/
/*END*/
